import threading


class cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.
    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


class thread_cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance per thread.
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__
        self.cache_name = "__cache_%s" % self.name

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        # Get threadlocal cache, making it if it doesn't exist
        if not hasattr(instance, self.cache_name):
            setattr(instance, self.cache_name, threading.local())
        cache = getattr(instance, self.cache_name)
        # Get cached result or make it
        if not hasattr(cache, "result"):
            cache.result = self.func(instance)
        return cache.result



from collections import deque


def dependency_sort(initial, dependencies):
    """
    Generic dependency sorting algorithm. Takes initial nodes, and a
    callable that returns a list of dependencies of a node given a node,
    and returns a list of the node and its dependencies from most depended
    (depends on nothing) to the node passed in (depends on everything else)
    """
    deps = list(initial)
    pending = deque(initial)
    mapping = {}
    while pending:
        current = pending.popleft()
        mapping[current] = [x for x in dependencies(current) if x is not None]
        for dep in mapping[current]:
            if dep not in pending and dep not in mapping:
                pending.append(dep)
    # Now roll through building a sorted list
    result = []
    while mapping:
        len_before = len(mapping)
        for node, deps in sorted(mapping.items()):
            if not deps or all((dep in result) for dep in deps):
                result.append(node)
                del mapping[node]
        if len(mapping) == len_before:
            raise ValueError("Circular dependency detected between: %s" % mapping.keys())
    return result



import attr
import click
import os
import subprocess
import sys

from .base import BasePlugin
from ..cli.argument_types import ContainerType, HostType
from ..cli.colors import RED
from ..docker.introspect import FormationIntrospector


@attr.s
class AttachPlugin(BasePlugin):
    """
    Plugin for attaching into a running container
    """

    def load(self):
        self.add_command(attach)


@click.command()
@click.argument("container", type=ContainerType())
@click.option("--host", "-h", type=HostType(), default="default")
@click.argument("command", nargs=-1, default=None)
@click.pass_obj
def attach(app, container, host, command):
    """
    Attaches to a container
    """
    if command:
        shell = ['/bin/bash', '-lc', ' '.join(command)]
    else:
        shell = ['/bin/bash']

    # See if the container is running
    formation = FormationIntrospector(host, app.containers).introspect()
    for instance in formation:
        if instance.container == container:
            # Work out anything to put before the shell (e.g. ENV)
            pre_args = []
            if os.environ.get("TERM", None):
                pre_args = ["env", "TERM=%s" % os.environ['TERM']]
            # Launch into an attached shell
            status_code = subprocess.call(["docker", "exec", "-it", instance.name] + pre_args + shell)
            sys.exit(status_code)
    # It's not running ;(
    click.echo(RED("Container {name} is not running. It must be started to attach - try `bay run {name}`.".format(
        name=container.name,
    )))



import attr
from click import Choice
from . import App
from .spell import spell_correct
from .colors import PURPLE, CYAN
from ..utils.functional import cached_property


Choice = attr.s(these={"_choices": attr.ib(default=None)}, init=False)(Choice)


@attr.s
class SpellCorrectChoice(Choice):
    # The superclass defines choices to be a required argument provided when
    # creating an instance of the class. Subclasses of Choice need to override
    # it so we can compute the choices as a cached property instead.
    context = attr.ib(init=False)

    def convert(self, value, param, ctx):
        self.context = ctx

        # Exact match
        if value in self.choices:
            return value

        # Match through normalization
        if ctx is not None and \
           ctx.token_normalize_func is not None:
            value = ctx.token_normalize_func(value)
            for choice in self.choices:
                if ctx.token_normalize_func(choice) == value:
                    return choice

        self.fail('invalid choice: %s. %s' % (PURPLE(value), self.get_missing_message(param, value)), param, ctx)

    def get_missing_message(self, param, value=None):
        message = super(SpellCorrectChoice, self).get_missing_message(param)
        if value:
            suggestion = spell_correct(value, self.choices)
            if suggestion:
                message += "\nare you looking for: {suggestion}?".format(suggestion=CYAN(suggestion))
        return message


@attr.s
class ContainerType(SpellCorrectChoice):
    name = 'container'
    _all = attr.ib(default=False)
    _profile = attr.ib(default=False)

    class Profile:
        pass

    @cached_property
    def choices(self):
        # Handle no object in the context during error states
        if not hasattr(self.context, "obj"):
            return App.get_default_containers().containers.keys()
        # Return valid choices
        containers = self.context.obj.containers
        choices = [container.name
                   for container in containers]
        if self._profile:
            choices = ['profile'] + choices
        if self._all:
            choices = ['all'] + choices
        return sorted(choices)

    def convert(self, value, param, ctx):
        name = super(ContainerType, self).convert(value, param, ctx)
        if name == 'all':
            return list(self.context.obj.containers.containers.values())
        elif name == 'profile':
            # Return a list of all containers
            return self.Profile
        else:
            return self.context.obj.containers[name]


@attr.s
class HostType(SpellCorrectChoice):
    name = 'host'

    @cached_property
    def choices(self):
        # Handle no object in the context during error states
        if not hasattr(self, "context") or not hasattr(self.context, "obj"):
            return []
        # Return valid choices
        choices = [host.alias for host in self.context.obj.hosts]
        return sorted(choices)

    def convert(self, value, param, ctx):
        host_name = super(HostType, self).convert(value, param, ctx)
        return self.context.obj.hosts[host_name]


@attr.s
class MountType(SpellCorrectChoice):
    name = 'mount'

    @cached_property
    def choices(self):
        # Handle no object in the context during error states
        if not hasattr(self.context, "obj"):
            return App.get_default_containers().devmode_names()
        # Collapse lists of list of devmode keys into a single set
        choices = self.context.obj.containers.devmode_names()

        return sorted(choices)



import attr
import click
from docker.errors import NotFound

from .base import BasePlugin
from ..cli.argument_types import HostType
from ..cli.tasks import Task
from ..exceptions import DockerRuntimeError
from ..utils.sorting import dependency_sort


@attr.s
class GcPlugin(BasePlugin):
    """
    Does garbage collection on demand.
    """

    provides = ["gc"]

    def load(self):
        self.add_command(gc)


@attr.s
class GarbageCollector:
    """
    Allows garbage collection on a host.
    """

    host = attr.ib()

    def gc_all(self, parent_task):
        task = Task("Running garbage collection", parent=parent_task)
        self.gc_containers(task)
        self.gc_remote_tags(task)
        self.gc_images(task)
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def gc_containers(self, parent_task):
        """
        Cleans up containers
        """
        task = Task("Removing dead containers", parent=parent_task)
        dead_containers = self.dead_containers()
        for i, name in enumerate(dead_containers):
            try:
                self.host.client.remove_container(name)
            except NotFound:
                raise DockerRuntimeError("Container {} vanished during gc".format(name))
            task.update(progress=(i + 1, len(dead_containers)))
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def dead_containers(self):
        """
        Gets all container IDs that are not actually running
        """
        live_containers = set(c['Id'] for c in self.host.client.containers(all=False, trunc=False, quiet=True))
        all_containers = set(c['Id'] for c in self.host.client.containers(all=True, trunc=False, quiet=True))
        return all_containers - live_containers

    def gc_remote_tags(self, parent_task):
        """
        Cleans up images without local tags (only remote tags)
        """
        task = Task("Removing remote tags", parent=parent_task)
        for image in self.host.client.images(all=False):
            for tag in (image.get("RepoTags", []) or []):
                if len(tag.split("/")) > 2:
                    self.host.client.remove_image(tag)
                    continue
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def gc_images(self, parent_task):
        """
        Cleans up old (untagged/non-current images)
        """
        # Clean up images
        task = Task("Removing dead images", parent=parent_task)

        # Get images with names and expand into all their layers
        named_images = self.named_images()
        current_images = self.image_layers(named_images)

        # Get all images used by running containers
        live_images = self.live_images()

        # Subtract all of those from a dependency sorted list of all images
        dead_images = [x for x in self.all_images() if x not in current_images and x not in live_images]

        # Delete those images in the presented (dependency) order
        for i, image_id in enumerate(dead_images):
            try:
                self.host.client.remove_image(image_id, force=True, noprune=True)
            except NotFound:
                raise DockerRuntimeError("Image {} vanished during gc".format(image_id))
            task.update(progress=(i + 1, len(dead_images)))
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def named_images(self):
        """
        Gets all images with a tag from the current set of Docker images.
        We can't use the config as a source as there might be images from
        other projects or docker tools.
        """
        images = set()
        for image in self.host.client.images(all=False):
            repo_tags = [
                tag for tag in (image.get("RepoTags", []) or [])
                if tag != "<none>:<none>" and len(tag.split("/")) < 3
            ]
            if repo_tags:
                images.add(image['Id'])
        return images

    def image_layers(self, named_images, force=False):
        """
        Gets image IDs of all layers used by the named images
        """
        current_images = set()
        for named_image in named_images:
            try:
                result = self.host.client.history(named_image)
            except NotFound:
                if force:
                    result = []
                else:
                    raise
            for image in result:
                current_images.add(image['Id'])
        return current_images

    def live_images(self, force=False):
        """
        Gets image IDs (including intermediate layers) that are used by running containers
        """
        live_images = set()
        for container in self.host.client.containers(all=False, trunc=False):
            try:
                result = self.host.client.history(container['Image'])
            except NotFound:
                if force:
                    result = []
                else:
                    raise
            for image in result:
                live_images.add(image['Id'])
        return live_images

    def all_images(self):
        """
        Returns all the live image IDs on the host sorted in dependency order, with
        the children first and ancestors last.
        """
        image_parents = {}
        for image in self.host.client.images(quiet=False, all=True):
            image_parents[image['Id']] = image['ParentId'] if image['ParentId'] else None
        all_images = dependency_sort(image_parents.keys(), lambda current: [image_parents[current]])
        all_images.reverse()
        return all_images


@click.command()
@click.option('--host', '-h', type=HostType(), default='default')
@click.pass_obj
def gc(app, host):
    """
    Runs the garbage collection manually.
    """
    GarbageCollector(host).gc_all(app.root_task)



import json
import tarfile
import time
from io import BytesIO

from docker.errors import NotFound


class Towline(object):
    """
    Process communication helper that can monitor the boot process and
    provide information about what's happening.
    """

    # Number of seconds till we conclude the container doesn't have towline support
    NO_TOWLINE_TIMEOUT = 2

    def __init__(self, host, container_name):
        self.host = host
        self.container_name = container_name
        self._first_try = None

    def _read_file(self, path, default=None):
        """
        Helper to read the contents of a file inside a container
        """
        try:
            tar_stream = self.host.client.get_archive(self.container_name, path)[0]
            tar = tarfile.open(fileobj=BytesIO(tar_stream.read()))
            contents = tar.extractfile(tar.getmembers()[0]).read().strip()
            return contents or default
        except NotFound:
            # Ignore missing containers or other errors
            return default

    @property
    def status(self):
        """
        Returns the container's current status as a (finished, message) tuple.
        Finished is True for successful boot, False for unsuccessful boot, and
        None if boot is still occuring.
        """
        # The container should exist by now
        if not self.host.container_exists(self.container_name):
            return (False, "Container does not exist")
        # If it's dead, that's a failed boot
        if not self.host.container_running(self.container_name):
            return (False, "Container died during boot")
        # See if we can read a status from it
        if self._first_try is None:
            self._first_try = time.time()
        container_status = self._read_file("/tugboat/boot_status")
        # If there's no status and the timeout has passed, they're not towline compatible
        if container_status is None and time.time() - self._first_try > self.NO_TOWLINE_TIMEOUT:
            return (True, "Non-towline boot complete")
        # See if boot is complete
        if self._read_file("/tugboat/boot_complete"):
            return (True, "Towline boot complete")
        elif container_status:
            # Try to parse out a JSON thing
            try:
                towline_payload = json.loads(container_status.split(b"\n")[-1].decode("ascii"))
                return (None, towline_payload['message'].rstrip(":"))
            except ValueError:
                return (None, container_status)
        else:
            return (None, None)



import attr
import click
import datetime
import sys
from docker.errors import NotFound

from .base import BasePlugin
from ..cli.colors import CYAN, GREEN, RED, remove_ansi
from ..cli.argument_types import ContainerType, HostType
from ..cli.tasks import Task
from ..constants import PluginHook
from ..docker.build import Builder
from ..docker.introspect import FormationIntrospector
from ..docker.runner import FormationRunner
from ..exceptions import BuildFailureError, ImagePullFailure
from .gc import GarbageCollector
from ..utils.sorting import dependency_sort


def _get_providers(app):
    providers = {}
    for container in app.containers:
        provides_volume = container.extra_data.get("provides-volume", None)
        if provides_volume:
            providers[provides_volume] = container
    return providers


def _handle_build_failure(app, logfile_name):
    click.echo(RED("Build failed! Last 15 lines of log:"))
    # TODO: More efficient tailing
    lines = []
    with open(logfile_name, "r") as fh:
        for line in fh:
            lines = lines[-14:] + [line]
    for line in lines:
        click.echo("  " + remove_ansi(line).rstrip())
    click.echo("See full build log at {log}".format(
        log=click.format_filename(logfile_name)),
        err=True
    )
    app.run_hooks(PluginHook.DOCKER_FAILURE)
    sys.exit(1)


@attr.s
class BuildPlugin(BasePlugin):
    """
    Plugin for building containers.
    """

    provides = ["build"]

    def load(self):
        self.add_command(build)
        self.add_catalog_type("registry")
        self.add_hook(PluginHook.PRE_START, self.pre_start)
        self.add_hook(PluginHook.POST_BUILD, self.post_build)

    def pre_start(self, host, instance, task):
        """
        Safety net to stop you booting volume-providing containers normally,
        and to catch and build volume containers if they're needed
        """
        # Safety net
        if instance.container.extra_data.get("provides-volume", None):
            raise ValueError("You cannot run a volume-providing container {}".format(instance.container.name))
        # If the container has named volumes, see if they're provided by anything else
        # and if so, if they're built.
        # First, collect what volumes are provided by what containers
        providers = _get_providers(self.app)
        # Now see if any of the volumes we're trying to add need it
        for _, volume in instance.container.named_volumes.items():
            name = volume.source
            if name in providers:
                # Alright, this is one that could be provided. Does it already exist?
                try:
                    host.client.inspect_volume(name)
                except NotFound:
                    # Aha! Build it!
                    try:
                        logfile_name = self.app.config.get_path(
                            'bay',
                            'build_log_path',
                            self.app,
                        )
                        Builder(
                            host,
                            providers[name],
                            self.app,
                            parent_task=task,
                            logfile_name=logfile_name,
                            verbose=True,
                        ).build()
                    except BuildFailureError:
                        _handle_build_failure(self.app, logfile_name)

    def post_build(self, host, container, task):
        """
        Intercepts builds of volume-providing containers and unpacks them.

        Volumes are stored with the ID of the corresponding volume-providing image. This will only run the container
        to recreate the volume if the image"s ID (hash) has changed.
        """
        image_details = host.client.inspect_image(container.image_name_tagged)
        provides_volume = container.extra_data.get("provides-volume", None)

        def should_extract_volume():
            if not provides_volume:
                return False
            try:
                volume_details = host.client.inspect_volume(provides_volume)
            except NotFound:
                return True
            labels = volume_details.get("Labels") or {}
            return labels.get("build_id") != image_details["Id"]

        if should_extract_volume():
            # Stop all containers that have the volume mounted
            formation = FormationIntrospector(host, self.app.containers).introspect()
            # Keep track of instances to remove after they are stopped
            instances_to_remove = formation.get_instances_using_volume(provides_volume)
            if instances_to_remove:
                formation.remove_instances(instances_to_remove)
                stop_task = Task("Stopping containers", parent=task)
                FormationRunner(self.app, host, formation, stop_task).run()
                stop_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
                remove_task = Task("Removing containers", parent=task)
                for instance in instances_to_remove:
                    host.client.remove_container(instance.name)
                    remove_task.update(status="Removed {}".format(instance.name))
                remove_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

            # Prune any orphan stopped containers, so we don't get conflict errors
            GarbageCollector(host).gc_containers(task)

            volume_task = Task("(Re)creating volume {}".format(provides_volume), parent=task)
            # Recreate the volume with the new image ID
            try:
                host.client.remove_volume(provides_volume)
                volume_task.update(status="Removed {}. Recreating".format(provides_volume))
            except NotFound:
                volume_task.update(status="Volume {} not found. Creating")
            host.client.create_volume(provides_volume, labels={"build_id": image_details["Id"]})

            # Configure the container
            volume_mountpoints = ["/volume/"]
            volume_binds = {provides_volume: {"bind": "/volume/", "mode": "rw"}}
            container_pointer = host.client.create_container(
                container.image_name,
                detach=False,
                volumes=volume_mountpoints,
                host_config=host.client.create_host_config(
                    binds=volume_binds,
                ),
            )
            # Start it in the foreground so we wait till it exits (detach=False above)
            volume_task.update(status="Extracting")
            host.client.start(container_pointer)
            host.client.wait(container_pointer["Id"])
            host.client.remove_container(container_pointer["Id"])
            volume_task.update(status="Done", status_flavor=Task.FLAVOR_GOOD)


@click.command()
@click.argument('containers', type=ContainerType(profile=True), nargs=-1)
@click.option('--host', '-h', type=HostType(), default='default')
@click.option('--cache/--no-cache', default=True)
@click.option('--recursive/--one', '-r/-1', default=True)
@click.option('--verbose/--quiet', '-v/-q', default=True)
# TODO: Add a proper requires_docker check
@click.pass_obj
def build(app, containers, host, cache, recursive, verbose):
    """
    Build container images, along with its build dependencies.
    """

    # `bay build` is equivalent to `bay build profile`
    if not containers:
        containers = [ContainerType.Profile]

    logfile_name = app.config.get_path('bay', 'build_log_path', app)
    containers_to_pull = []
    containers_to_build = []
    pulled_containers = set()
    failed_pulls = set()

    task = Task("Building", parent=app.root_task)
    start_time = datetime.datetime.now().replace(microsecond=0)

    providers = _get_providers(app)

    # Go through the containers, expanding "ContainerType.Profile" into a list
    # of default boot containers in the profile.
    for container in containers:
        if container is ContainerType.Profile:
            for con in app.containers:
                # When building the profile, rebuild system containers too
                if app.containers.options(con).get('in_profile') or con.system:
                    containers_to_pull.append(con)
        else:
            containers_to_build.append(container)
            for volume in container.named_volumes.values():
                if volume.source in providers:
                    containers_to_build.append(providers[volume.source])

    # Expand containers_to_pull (At this point just the default boot containers
    # from profile) to include runtime dependencies.
    containers_to_pull = dependency_sort(containers_to_pull, app.containers.dependencies)

    # Expand containers_to_pull to include volumes that are required by all containers in the
    # dependency chain.
    def container_volume_dependencies(container):
        volume_deps = set()
        for volume in container.named_volumes.values():
            if volume.source in providers:
                volume_deps.add(providers[volume.source])
        return volume_deps

    containers_to_pull = dependency_sort(containers_to_pull, container_volume_dependencies)

    # Try pulling each container to pull, and add it to containers_to_build if
    # it fails. If it works, remember we pulled it, so we don't have to pull it
    # again later.
    for container in containers_to_pull:
        try:
            host.images.pull_image_version(
                app,
                container.image_name,
                container.image_tag,
                parent_task=task,
                fail_silently=False,
            )
        except ImagePullFailure:
            failed_pulls.add(container)
            containers_to_build.append(container)
        else:
            pulled_containers.add(container)

    ancestors_to_build = []
    # For each container to build, find its ancestry, trying to pull each
    # ancestor and stopping short if it works.
    for container in containers_to_build:
        # Always add `container` to final build list, even if recursive is
        # False.
        ancestors_to_build.append(container)
        if recursive:
            # We need to look at the ancestry starting from the oldest, up to
            # and not including the `container`
            ancestry = app.containers.build_ancestry(container)
            for ancestor in reversed(ancestry):
                try:
                    # If we've already attempted to pull it and failed, short
                    # circuit to failure block.
                    if ancestor in failed_pulls:
                        raise ImagePullFailure("We've already attempted to pull this image, and failed.")
                    # Check if we've pulled it already
                    if ancestor not in pulled_containers:
                        host.images.pull_image_version(
                            app,
                            ancestor.image_name,
                            ancestor.image_tag,
                            parent_task=task,
                            fail_silently=False,
                        )
                except ImagePullFailure:
                    failed_pulls.add(ancestor)
                    ancestors_to_build.insert(0, ancestor)
                else:
                    # We've pulled the current ancestor successfully, so skip
                    # all the older ancestors.
                    pulled_containers.add(ancestor)
                    break

    # Sort ancestors so we build the most depended on first.
    sorted_ancestors_to_build = dependency_sort(ancestors_to_build,
                                                lambda x: [app.containers.build_parent(x)])

    # dependency_sort would insert back the pulled containers into the ancestry
    # chain, so we only include ones that were in the list before
    ancestors_to_build = [
        container
        for container in sorted_ancestors_to_build
        if container in ancestors_to_build
    ]

    task.add_extra_info(
        "Order: {order}".format(
            order=CYAN(", ".join([container.name for container in ancestors_to_build])),
        ),
    )

    app.run_hooks(PluginHook.PRE_GROUP_BUILD, host=host, containers=ancestors_to_build, task=task)

    for container in ancestors_to_build:
        image_builder = Builder(
            host,
            container,
            app,
            parent_task=task,
            logfile_name=logfile_name,
            docker_cache=cache,
            verbose=verbose,
        )
        try:
            image_builder.build()
        except BuildFailureError:
            app.run_hooks(PluginHook.CONTAINER_FAILURE, host=host, containers=ancestors_to_build, task=task)
            _handle_build_failure(app, logfile_name)

    app.run_hooks(PluginHook.POST_GROUP_BUILD, host=host, containers=ancestors_to_build, task=task)

    task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    # Show total build time metric after everything is complete
    end_time = datetime.datetime.now().replace(microsecond=0)
    time_delta_str = str(end_time - start_time)
    if time_delta_str.startswith('0:'):
        # no point in showing hours, unless it runs for more than one hour
        time_delta_str = time_delta_str[2:]
    click.echo("Total build time [{}]".format(GREEN(time_delta_str)))



import attr
import click

from .base import BasePlugin
from ..cli.argument_types import ContainerType
from ..cli.colors import CYAN
from ..cli.table import Table


@attr.s
class ContainerPlugin(BasePlugin):
    """
    Plugin for showing information about containers
    """

    def load(self):
        self.add_command(container)


@click.command()
@click.argument("container", type=ContainerType(), required=False)
@click.pass_obj
def container(app, container=None):
    """
    Shows details on containers
    """
    if container is None:
        # Print containers
        table = Table([
            ("NAME", 30),
        ])
        table.print_header()
        for container in sorted(app.containers, key=lambda c: c.name):
            table.print_row([
                container.name,
            ])
    else:
        # Container name
        click.echo(CYAN("Name: ") + container.name)
        # Build parent
        click.echo(
            CYAN("Build ancestry: ") +
            ", ".join(other.name for other in app.containers.build_ancestry(container))
        )
        # Runtime dependencies
        dependencies = app.containers.dependencies(container)
        if dependencies:
            click.echo(CYAN("Depends on: ") + ", ".join(sorted(other.name for other in dependencies)))
        else:
            click.echo(CYAN("Depends on: ") + "(nothing)")
        # Dependents
        dependents = app.containers.dependents(container)
        if dependents:
            click.echo(CYAN("Depended on by: ") + ", ".join(sorted(other.name for other in dependents)))
        else:
            click.echo(CYAN("Depended on by: ") + "(nothing)")
        # Volumes
        click.echo(CYAN("Named volumes:"))
        for mount_point, volume in container.named_volumes.items():
            click.echo("  {}: {}".format(mount_point, volume.source))
        click.echo(CYAN("Bind-mounted volumes:"))
        for mount_point, volume in container.bound_volumes.items():
            click.echo("  {}: {}".format(mount_point, volume.source))
        # Devmodes
        click.echo(CYAN("Mounts (devmodes):"))
        for name, mounts in container.devmodes.items():
            click.echo("  {}:".format(name))
            for mount_point, volume in mounts.items():
                click.echo("    {}: {}".format(mount_point, volume.source))



from .base import BasePlugin
from ..constants import PluginHook
from ..docker.introspect import FormationIntrospector


class SystemContainerBuildPlugin(BasePlugin):

    def load(self):
        self.system_contaienr_cache = {}
        self.add_hook(PluginHook.POST_GROUP_BUILD, self.post_group_build)

    def post_group_build(self, host, containers, task):
        """Restart all running system containers whose IDs have changed."""
        formation = FormationIntrospector(host, self.app.containers).introspect()
        containers_to_restart = set()
        for container in containers:
            if container.system:
                # If the running instance is based on an outdated image, restart it
                try:
                    instance = formation.get_container_instance(container.name)
                except ValueError:
                    continue
                image_details = host.client.inspect_image(container.image_name)
                if instance and image_details and image_details["Id"] != instance.image_id:
                    containers_to_restart.add(container)
        if containers_to_restart:
            self.app.invoke("restart", containers=containers_to_restart)



from setuptools import setup
from bay import __version__

setup(
    name='bay',
    version=__version__,
    author='Eventbrite, Inc.',
    description='Docker-based development workflow tooling',
    long_description=(
        'Bay is a tool for assisting creation and management of Docker ' +
        'containers for development use.  It allows you to supplement a ' +
        'Dockerfile with additional information on how to run and link ' +
        'containers together. \n\nFor more, see http://github.com/eventbrite/bay/'
    ),
    packages=[
        "bay",
        "bay.cli",
        "bay.containers",
        "bay.docker",
        "bay.plugins",
        "bay.utils",
    ],
    include_package_data=True,
    install_requires=[
        'attrs',
        'Click>=6.6',
        'PyYAML',
        'docker~=2.5.1',
        'dockerpty==0.4.1',
        'scandir',
        'requests<2.11.0',
    ],
    extras_require={
        'spell': ['pylev'],
    },
    test_suite="tests",
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    entry_points='''
        [console_scripts]
        bay = bay.cli:cli
        tug = bay.cli:cli

        [bay.plugins]
        attach = bay.plugins.attach:AttachPlugin
        boot = bay.plugins.boot:BootPlugin
        build = bay.plugins.build:BuildPlugin
        build_scripts = bay.plugins.build_scripts:BuildScriptsPlugin
        container = bay.plugins.container:ContainerPlugin
        doctor = bay.plugins.doctor:DoctorPlugin
        gc = bay.plugins.gc:GcPlugin
        help = bay.plugins.help:HelpPlugin
        hosts = bay.plugins.hosts:HostsPlugin
        images = bay.plugins.images:ImagesPlugin
        legacy_env = bay.plugins.legacy_env:LegacyEnvPlugin
        mounts = bay.plugins.mounts:DevModesPlugin
        profile = bay.plugins.profile:ProfilesPlugin
        ps = bay.plugins.ps:PsPlugin
        registry = bay.plugins.registry:RegistryPlugin
        run = bay.plugins.run:RunPlugin
        ssh_agent = bay.plugins.ssh_agent:SSHAgentPlugin
        system = bay.plugins.system:SystemContainerBuildPlugin
        tail = bay.plugins.tail:TailPlugin
        volume = bay.plugins.volume:VolumePlugin
        waits = bay.plugins.waits:WaitsPlugin
    ''',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
    ],
)



import os
import sys
import base64
import click
import subprocess

from .base import BasePlugin
from ..docker.introspect import FormationIntrospector
from ..exceptions import DockerRuntimeError
from ..constants import PluginHook


class SSHAgentPlugin(BasePlugin):
    """
    Plugin for running an SSH agent container at boot.

    This container needs special handling as it needs local keys injected into
    environment variables at runtime.
    """

    CONTAINER_NAME = "ssh-agent"

    requires = ["boot-containers"]

    def load(self):
        self.add_hook(PluginHook.PRE_BUILD, self.pre_build)
        self.add_hook(PluginHook.PRE_START, self.pre_start)

    def pre_build(self, host, container, task):
        """
        Injects the SSH_AUTH_HOST variable into builds if it's needed.
        """
        # TODO: Rename this to SSH_AUTH_HOST
        # See if this container needs ssh-agent during build
        uses_ssh, required = self.container_needs_ssh(container, "build")
        if not uses_ssh:
            return
        # See if the SSH auth container is running
        if self.ssh_container_running(host):
            if 'SSH_AUTH_HOST' in container.possible_buildargs:
                container.buildargs['SSH_AUTH_HOST'] = host.build_host_ip
            # TODO Deprecate TUGBOAT_SSH_AUTH_HOST by deleting the below lines.
            if 'TUGBOAT_SSH_AUTH_HOST' in container.possible_buildargs:
                container.buildargs['TUGBOAT_SSH_AUTH_HOST'] = host.build_host_ip
        elif required:
            raise DockerRuntimeError(
                "The container {} needs an SSH Agent to build but one is not started".format(container.name),
            )

    def pre_start(self, host, instance, task):
        """
        Catches the ssh-agent container being started and gives it the right
        environment variables, as well as passing the auth details into containers.
        """
        # See if it's the ssh-agent container that we need to give runtime config to
        if instance.container.name == self.CONTAINER_NAME:
            for i, path in enumerate(self.key_paths()):
                instance.environment["SSHKEY%s" % i] = self.read_key(path)
                task.add_extra_info("Key {}: {}".format(i, path))
        # Or another container that needs the env variable
        else:
            # See if this container needs ssh-agent during build
            uses_ssh, required = self.container_needs_ssh(instance.container, "run")
            if not uses_ssh:
                return
            # See if the SSH auth container is running
            if self.ssh_container_running(host):
                instance.environment['SSH_AUTH_HOST'] = host.build_host_ip
                # TODO Deprecate TUGBOAT_SSH_AUTH_HOST by deleting the below line.
                instance.environment['TUGBOAT_SSH_AUTH_HOST'] = host.build_host_ip
            elif required:
                raise DockerRuntimeError(
                    "The container {} needs an SSH Agent to run but one is not started".format(instance.container.name)
                )

    def container_needs_ssh(self, container, phase):
        """
        Returns (True, True) if the container needs SSH to start,
        (True, False) if it can use it but doesn't require it, or
        (False, False) if it doesn't need it.
        """
        configs = container.get_ancestral_extra_data("boot")
        uses_ssh = False
        required = False
        for config in configs:
            for name, required in config.get(phase, {}).items():
                if name == self.CONTAINER_NAME:
                    required = required.lower().strip() == "required"
                    uses_ssh = True
        return uses_ssh, required

    def ssh_container_running(self, host):
        """
        Returns True if the agent container is running on the host, False otherwise.
        """
        formation = FormationIntrospector(host, self.app.containers).introspect()
        for instance in formation:
            if instance.container.name == self.CONTAINER_NAME:
                return True
        return False

    def key_paths(self):
        """
        Returns paths to the current user's SSH keys
        """
        options = [
            "~/.ssh/id_ecdsa",
            "~/.ssh/id_rsa",
            "~/.ssh/id_dsa",
            "~/.ssh/identity",
        ]
        for option in options:
            path = os.path.expanduser(option)
            if os.path.isfile(path):
                yield path

    def read_key(self, path):
        """
        Returns a key, optionally prompting for passphrase.
        """
        # Read key
        with open(path, "rb") as fh:
            key_contents = fh.read()
        # If it's encrypted, then decrypt it
        if b"ENCRYPTED" in key_contents:
            with self.app.root_task.paused_output():
                print("Decrypting key %s" % path)
                rsa_proc = subprocess.Popen(
                    ["openssl", "rsa"],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                key_contents = rsa_proc.communicate(input=key_contents)[0]
                if rsa_proc.returncode:
                    click.echo(click.style(
                        "Incorrect passphrase!",
                        fg="red",
                    ))
                    sys.exit(1)
        # Return key b64 encoded
        return base64.b64encode(key_contents)
